import cv2
import numpy as np
from skimage import transform as trans
from PIL import Image


def glint360k_align(img, landmark, output_size=(112,112), pil_output=True):
    """
 
    Args:
        img: PIL Image or cv2 img(np array)
        landmark: faceid landmarks [[x1,y1],[x2,y2],..[x5,y5]] for [lefteye, righteye, nose, mouthleft, mouthright]
        output_size: output Image size
        pil_output: True return PIL Image, False return cv2 image

    Returns:
        A Image with PIL or cv2 format

    Raises:
        IOError: img is not PIL or cv2 Image
     
    """

    if type(img) is not np.ndarray:
        try:
            img = np.asarray(img)
        except:
            raise "alignment should input an ndarray or Image"

    dst = np.array([
   [30.2946, 51.6963],
   [65.5318, 51.5014],
   [48.0252, 71.7366],
   [33.5493, 92.3655],
   [62.7299, 92.2041]], dtype=np.float32 )
    dst[:,0] += 8.0
    src = landmark.astype(np.float32)
    tform = trans.SimilarityTransform()
    tform.estimate(src, dst)
    M = tform.params[0:2, :]
    img = cv2.warpAffine(img, M, output_size, borderValue=0.0)

    if pil_output:
        img = Image.fromarray(img)

    return img


# Copyright (c) Alibaba, Inc. and its affiliates.
import math
import os

import numpy as np
import torch
from PIL import Image, ImageFile

from easycv.file import io
from easycv.framework.errors import ValueError
from easycv.utils.misc import deprecated
from .base import InputProcessor, OutputProcessor, Predictor, PredictorV2
from .builder import PREDICTORS
from .classifier import ClassificationPredictor


@PREDICTORS.register_module()
class ReIDPredictor(ClassificationPredictor):
    """Predictor for reid.
    Args:
        model_path (str): Path of model path.
        config_file (Optinal[str]): config file path for model and processor to init. Defaults to None.
        batch_size (int): batch size for forward.
        device (str): Support 'cuda' or 'cpu', if is None, detect device automatically.
        save_results (bool): Whether to save predict results.
        save_path (str): File path for saving results, only valid when `save_results` is True.
        pipelines (list[dict]): Data pipeline configs.
        topk (int): Return top-k results. Default: 1.
        pil_input (bool): Whether use PIL image. If processor need PIL input, set true, default false.
        label_map_path (str): File path of saving labels list.
        input_processor_threads (int): Number of processes to process inputs.
        mode (str): The image mode into the model.
    """

    def __init__(self,
                 model_path,
                 config_file=None,
                 batch_size=1,
                 device=None,
                 save_results=False,
                 save_path=None,
                 pipelines=None,
                 topk=1,
                 pil_input=True,
                 label_map_path=None,
                 input_processor_threads=8,
                 mode='BGR',
                 *args,
                 **kwargs):

        if pil_input:
            mode = 'RGB'
        super(ReIDPredictor, self).__init__(
            model_path,
            config_file=config_file,
            batch_size=batch_size,
            device=device,
            save_results=save_results,
            save_path=save_path,
            pipelines=pipelines,
            input_processor_threads=input_processor_threads,
            mode=mode,
            topk=topk,
            pil_input=pil_input,
            label_map_path=label_map_path,
            *args,
            **kwargs)

    def model_forward(self, inputs):
        """Model forward.
        If you need refactor model forward, you need to reimplement it.
        """
        with torch.no_grad():
            outputs = self.model(**inputs, mode='extract')
        return outputs

    def get_id(self, img_path):
        camera_id = []
        labels = []
        for path in img_path:
            filename = os.path.basename(path)
            label = filename[0:4]
            camera = filename.split('c')[1]
            if label[0:2] == '-1':
                labels.append(-1)
            else:
                labels.append(int(label))
            camera_id.append(int(camera[0]))
        return camera_id, labels

    def file_name_walk(self, file_dir):
        image_path_list = []
        for root, dirs, files in os.walk(file_dir):
            for name in files:
                if name.endswith('.jpg'):
                    image_path_list.append(os.path.join(root, name))
        return image_path_list

    def __call__(self, inputs, keep_inputs=False):
        if not isinstance(inputs, list):
            inputs = self.file_name_walk(inputs)
            img_cam, img_label = self.get_id(inputs)
        else:
            img_cam, img_label = None, None

        if self.input_processor is None:
            self.input_processor = self.get_input_processor()

        # TODO: fault tolerance
        if isinstance(inputs, (str, np.ndarray, ImageFile.ImageFile)):
            inputs = [inputs]

        results_list = []
        for i in range(0, len(inputs), self.batch_size):
            batch_inputs = inputs[i:min(len(inputs), i + self.batch_size)]
            batch_outputs = self.input_processor(batch_inputs)
            batch_outputs = self._to_device(batch_outputs)
            batch_outputs = self.model_forward(batch_outputs)
            results = batch_outputs['neck']
            if keep_inputs:
                for i in range(len(batch_inputs)):
                    results[i].update({'inputs': batch_inputs[i]})
            if isinstance(results, list):
                results_list.extend(results)
            else:
                results_list.append(results)

        image_feature = torch.cat(results_list, 0)
        image_feature_norm = torch.norm(
            image_feature, p=2, dim=1, keepdim=True)
        image_feature = image_feature.div(
            image_feature_norm.expand_as(image_feature))

        # TODO: support append to file
        if self.save_results:
            self.dump(image_feature, self.save_path)

        return {
            'img_feature': image_feature,
            'img_cam': img_cam,
            'img_label': img_label
        }



# Copyright 2017 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Class for evaluating object detections with COCO metrics."""
from __future__ import print_function
import json
import multiprocessing
import os
import tempfile
from collections import OrderedDict, defaultdict

import mmcv
import numpy as np
import six
import torch
from xtcocotools.coco import COCO
from xtcocotools.cocoeval import COCOeval

from easycv.core import standard_fields
from easycv.core.evaluation import coco_tools
from easycv.core.post_processing.nms import oks_nms, soft_oks_nms
from easycv.core.standard_fields import DetectionResultFields, InputDataFields
from easycv.framework.errors import KeyError, TypeError, ValueError
from easycv.utils.json_utils import MyEncoder
from .base_evaluator import Evaluator
from .builder import EVALUATORS
from .metric_registry import METRICS

try:
    from panopticapi.evaluation import OFFSET, VOID, PQStat
    from panopticapi.utils import rgb2id
except ImportError:
    PQStat = None
    rgb2id = None
    VOID = 0
    OFFSET = 256 * 256 * 256


@EVALUATORS.register_module
class CocoDetectionEvaluator(Evaluator):
    """Class to evaluate COCO detection metrics."""

    def __init__(self,
                 classes,
                 include_metrics_per_category=False,
                 all_metrics_per_category=False,
                 coco_analyze=False,
                 dataset_name=None,
                 metric_names=['DetectionBoxes_Precision/mAP']):
        """Constructor.

        Args:
            classes: a list of class name
            include_metrics_per_category: If True, include metrics for each category.
            all_metrics_per_category: Whether to include all the summary metrics for
                each category in per_category_ap. Be careful with setting it to true if
                you have more than handful of ∏, because it will pollute your mldash.
            coco_analyze: If True, will analyze the detection result using coco analysis.
            dataset_name: If not None, dataset_name will be inserted to each metric name.
        """

        super(CocoDetectionEvaluator, self).__init__(dataset_name,
                                                     metric_names)
        # _image_ids is a dictionary that maps unique image ids to Booleans which
        # indicate whether a corresponding detection has been added.
        self._image_ids = {}
        self._groundtruth_list = []
        self._detection_boxes_list = []
        self._annotation_id = 1
        self._metrics = None
        self._analyze_images = None
        self._include_metrics_per_category = include_metrics_per_category
        self._all_metrics_per_category = all_metrics_per_category
        self._analyze = coco_analyze
        self._categories = [None] * len(classes)

        # categories: A list of dicts, each of which has the following keys -
        #    'id': (required) an integer id uniquely identifying this category.
        #    'name': (required) string representing category name e.g., 'cat', 'dog'.
        for idx, name in enumerate(classes):
            self._categories[idx] = {'id': idx, 'name': name}
        self._category_id_set = set([cat['id'] for cat in self._categories])

        # for json formatted evaluation
        self.num_classes = len(classes)
        self.class_names = classes
        self.dataset_name = dataset_name
        self.iou_thrs = np.linspace(
            .5, 0.95, int(np.round((0.95 - .5) / .05)) + 1, endpoint=True)
        self.proposal_nums = (100, 300, 1000)
        self.coco_metric_names = {
            'mAP': 0,
            'mAP_50': 1,
            'mAP_75': 2,
            'mAP_s': 3,
            'mAP_m': 4,
            'mAP_l': 5,
            'AR@100': 6,
            'AR@300': 7,
            'AR@1000': 8,
            'AR_s@1000': 9,
            'AR_m@1000': 10,
            'AR_l@1000': 11
        }
        self.metric_items = [
            'mAP', 'mAP_50', 'mAP_75', 'mAP_s', 'mAP_m', 'mAP_l'
        ]
        self.metric = 'bbox'
        self.iou_type = 'bbox'

    def clear(self):
        """Clears the state to prepare for a fresh evaluation."""
        self._image_ids.clear()
        self._groundtruth_list = []
        self._detection_boxes_list = []

    def add_single_ground_truth_image_info(self, image_id, groundtruth_dict):
        """Adds groundtruth for a single image to be used for evaluation.

        If the image has already been added, a warning is logged, and groundtruth is
        ignored.

        Args:
            image_id: A unique string/integer identifier for the image.
            groundtruth_dict: A dictionary containing

                :InputDataFields.groundtruth_boxes: float32 numpy array of shape
                    [num_boxes, 4] containing `num_boxes` groundtruth boxes of the format
                    [ymin, xmin, ymax, xmax] in absolute image coordinates.
                :InputDataFields.groundtruth_classes: integer numpy array of shape
                    [num_boxes] containing 1-indexed groundtruth classes for the boxes.
                    InputDataFields.groundtruth_is_crowd (optional): integer numpy array of
                    shape [num_boxes] containing iscrowd flag for groundtruth boxes.
        """
        if image_id in self._image_ids:
            print(
                'Ignoring ground truth with image id %s since it was '
                '  previously added', image_id)
            return

        groundtruth_is_crowd = groundtruth_dict.get(
            standard_fields.InputDataFields.groundtruth_is_crowd)
        # Drop groundtruth_is_crowd if empty tensor.
        if groundtruth_is_crowd is not None and not groundtruth_is_crowd.shape[
                0]:
            groundtruth_is_crowd = None

        self._groundtruth_list.extend(
            coco_tools.ExportSingleImageGroundtruthToCoco(
                image_id=image_id,
                next_annotation_id=self._annotation_id,
                category_id_set=self._category_id_set,
                groundtruth_boxes=groundtruth_dict[
                    standard_fields.InputDataFields.groundtruth_boxes],
                groundtruth_classes=groundtruth_dict[
                    standard_fields.InputDataFields.groundtruth_classes],
                groundtruth_is_crowd=groundtruth_is_crowd))
        self._annotation_id += groundtruth_dict[
            standard_fields.InputDataFields.groundtruth_boxes].shape[0]
        # Boolean to indicate whether a detection has been added for this image.
        self._image_ids[image_id] = False

    def add_single_detected_image_info(self, image_id, detections_dict):
        """Adds detections for a single image to be used for evaluation.

        If a detection has already been added for this image id, a warning is
        logged, and the detection is skipped.

        Args:
            image_id: A unique string/integer identifier for the image.
            detections_dict: A dictionary containing

                :DetectionResultFields.detection_boxes: float32 numpy array of shape
                    [num_boxes, 4] containing `num_boxes` detection boxes of the format
                    [ymin, xmin, ymax, xmax] in absolute image coordinates.
                :DetectionResultFields.detection_scores: float32 numpy array of shape
                    [num_boxes] containing detection scores for the boxes.
                :DetectionResultFields.detection_classes: integer numpy array of shape
                    [num_boxes] containing 1-indexed detection classes for the boxes.

        Raises:
            ValueError: If groundtruth for the image_id is not available.
        """
        if image_id not in self._image_ids:
            raise ValueError(
                'Missing groundtruth for image id: {}'.format(image_id))

        if self._image_ids[image_id]:
            print(
                'Ignoring detection with image id %s since it was '
                'previously added', image_id)
            return

        self._detection_boxes_list.extend(
            coco_tools.ExportSingleImageDetectionBoxesToCoco(
                image_id=image_id,
                category_id_set=self._category_id_set,
                detection_boxes=detections_dict[
                    standard_fields.DetectionResultFields.detection_boxes],
                detection_scores=detections_dict[
                    standard_fields.DetectionResultFields.detection_scores],
                detection_classes=detections_dict[
                    standard_fields.DetectionResultFields.detection_classes]))
        self._image_ids[image_id] = True

    def _evaluate(self, analyze=False):
        """Evaluates the detection boxes and returns a dictionary of coco metrics.
        Args:
            analyze:  if set True, will call coco analyze to analyze false positive,
                return result images for each class, this process is very slow.

        Returns:
            A dictionary holding

            1. summary_metrics:
                'DetectionBoxes_Precision/mAP': mean average precision over classes
                    averaged over IOU thresholds ranging from .5 to .95 with .05
                    increments.
                'DetectionBoxes_Precision/mAP@.50IOU': mean average precision at 50% IOU
                'DetectionBoxes_Precision/mAP@.75IOU': mean average precision at 75% IOU
                'DetectionBoxes_Precision/mAP (small)': mean average precision for small
                    objects (area < 32^2 pixels).
                'DetectionBoxes_Precision/mAP (medium)': mean average precision for
                    medium sized objects (32^2 pixels < area < 96^2 pixels).
                'DetectionBoxes_Precision/mAP (large)': mean average precision for large
                    objects (96^2 pixels < area < 10000^2 pixels).
                'DetectionBoxes_Recall/AR@1': average recall with 1 detection.
                'DetectionBoxes_Recall/AR@10': average recall with 10 detections.
                'DetectionBoxes_Recall/AR@100': average recall with 100 detections.
                'DetectionBoxes_Recall/AR@100 (small)': average recall for small objects
                    with 100.
                'DetectionBoxes_Recall/AR@100 (medium)': average recall for medium objects
                    with 100.
                'DetectionBoxes_Recall/AR@100 (large)': average recall for large objects
                    with 100 detections.

            2. per_category_ap: if include_metrics_per_category is True, category
            specific results with keys of the form:
                'Precision mAP ByCategory/category' (without the supercategory part if
                no supercategories exist). For backward compatibility
                'PerformanceByCategory' is included in the output regardless of
                all_metrics_per_category.

        """
        groundtruth_dict = {
            'annotations': self._groundtruth_list,
            'images': [{
                'id': image_id
            } for image_id in self._image_ids],
            'categories': self._categories
        }
        coco_wrapped_groundtruth = coco_tools.COCOWrapper(groundtruth_dict)
        coco_wrapped_detections = coco_wrapped_groundtruth.LoadAnnotations(
            self._detection_boxes_list)
        box_evaluator = coco_tools.COCOEvalWrapper(
            coco_wrapped_groundtruth,
            coco_wrapped_detections,
            agnostic_mode=False)
        box_metrics, box_per_category_ap = box_evaluator.ComputeMetrics(
            include_metrics_per_category=self._include_metrics_per_category,
            all_metrics_per_category=self._all_metrics_per_category)
        box_metrics.update(box_per_category_ap)
        box_metrics = {
            'DetectionBoxes_' + key: value
            for key, value in iter(box_metrics.items())
        }

        if self._analyze or analyze:
            images = box_evaluator.Analyze()
            return box_metrics, images
        else:
            return box_metrics

    def _evaluate_impl(self, prediction_dict, groundtruth_dict):
        '''
        Args:
            prediction_dict:  A dict of k-v pair, each v is a list of
                tensor or numpy array for detection result. A dictionary containing
                :groundtruth_boxes: float32 numpy array of shape
                    [num_boxes, 4] containing `num_boxes` groundtruth boxes of the format
                    [ymin, xmin, ymax, xmax] in absolute image coordinates.
                :groundtruth_classes: integer numpy array of shape
                    [num_boxes] containing 1-indexed groundtruth classes for the boxes.
                    InputDataFields.groundtruth_is_crowd (optional): integer numpy array of
                    shape [num_boxes] containing iscrowd flag for groundtruth boxes.
                :img_metas: List of length number of test images,
                        dict of image meta info, containing filename, ori_img_shape, and so on.
            groundtruth_dict: A dict of k-v pair, each v is a list of
                tensor or numpy array for groundtruth info. A dictionary containing
                :DetectionResultFields.detection_boxes: float32 numpy array of shape
                    [num_boxes, 4] containing `num_boxes` detection boxes of the format
                    [ymin, xmin, ymax, xmax] in absolute image coordinates.
                :DetectionResultFields.detection_scores: float32 numpy array of shape
                    [num_boxes] containing detection scores for the boxes.
                :DetectionResultFields.detection_classes: integer numpy array of shape
                    [num_boxes] containing 1-indexed detection classes for the boxes.
                :groundtruth_is_crowd: integer numpy array of
                    shape [num_boxes] containing iscrowd flag for groundtruth boxes.

        Return:
            dict,  each key is metric_name, value is metric value
        '''
        num_images_det = len(
            prediction_dict[DetectionResultFields.detection_boxes])
        num_images_gt = len(
            groundtruth_dict[InputDataFields.groundtruth_boxes])
        assert num_images_det == num_images_gt, 'detection and groundtruth number is not the same'

        groundtruth_is_crowd_list = groundtruth_dict.get(
            'groundtruth_is_crowd', None)
        for idx, (detection_boxes, detection_classes, detection_scores, img_metas, gt_boxes, gt_classes) in \
            enumerate(zip(prediction_dict['detection_boxes'], prediction_dict['detection_classes'],
                          prediction_dict['detection_scores'], prediction_dict['img_metas'],
                          groundtruth_dict['groundtruth_boxes'], groundtruth_dict['groundtruth_classes'])):

            height, width = img_metas[
                'ori_img_shape'][:2]  # height, width, channel
            image_id = img_metas['filename']
            # tensor caculated on other devices is not on device:0, so transfer them
            if isinstance(detection_boxes, torch.Tensor):
                detection_boxes = detection_boxes.cpu().numpy()
                detection_scores = detection_scores.cpu().numpy()
                detection_classes = detection_classes.cpu().numpy().astype(
                    np.int32)

            if groundtruth_is_crowd_list is None:
                groundtruth_is_crowd = None
            else:
                groundtruth_is_crowd = groundtruth_is_crowd_list[idx]

            groundtruth_dict = {
                'groundtruth_boxes': gt_boxes,
                'groundtruth_classes': gt_classes,
                'groundtruth_is_crowd': groundtruth_is_crowd,
            }
            self.add_single_ground_truth_image_info(image_id, groundtruth_dict)
            if detection_classes is None or detection_scores is None:
                detection_classes = np.array([-1])
                detection_scores = np.array([0.0])
                detection_boxes = np.array([[0.0, 0.0, 0.0, 0.0]])

            # add detection info
            detection_dict = {
                'detection_boxes': detection_boxes,
                'detection_scores': detection_scores,
                'detection_classes': detection_classes,
            }
            self.add_single_detected_image_info(image_id, detection_dict)

        eval_dict = self._evaluate()
        self.clear()
        return eval_dict


def _check_mask_type_and_value(array_name, masks):
    """Checks whether mask dtype is uint8 and the values are either 0 or 1."""
    if masks.dtype != np.uint8:
        raise TypeError('{} must be of type np.uint8. Found {}.'.format(
            array_name, masks.dtype))
    if np.any(np.logical_and(masks != 0, masks != 1)):
        raise ValueError(
            '{} elements can only be either 0 or 1.'.format(array_name))


@EVALUATORS.register_module
class CocoMaskEvaluator(Evaluator):
    """Class to evaluate COCO detection metrics."""

    def __init__(self,
                 classes,
                 include_metrics_per_category=False,
                 dataset_name=None,
                 metric_names=['DetectionMasks_Precision/mAP']):
        """Constructor.

        Args:
            categories: A list of dicts, each of which has the following keys
                :id: (required) an integer id uniquely identifying this category.
                :name: (required) string representing category name e.g., 'cat', 'dog'.
            include_metrics_per_category: If True, include metrics for each category.
        """
        super(CocoMaskEvaluator, self).__init__(dataset_name, metric_names)
        self._image_id_to_mask_shape_map = {}
        self._image_ids_with_detections = set([])
        self._groundtruth_list = []
        self._detection_masks_list = []
        self._annotation_id = 1
        self._include_metrics_per_category = include_metrics_per_category
        self._categories = [None] * len(classes)

        # categories: A list of dicts, each of which has the following keys -
        #    'id': (required) an integer id uniquely identifying this category.
        #    'name': (required) string representing category name e.g., 'cat', 'dog'.
        for idx, name in enumerate(classes):
            self._categories[idx] = {'id': idx, 'name': name}
        self._category_id_set = set([cat['id'] for cat in self._categories])

    def clear(self):
        """Clears the state to prepare for a fresh evaluation."""
        self._image_id_to_mask_shape_map.clear()
        self._image_ids_with_detections.clear()
        self._groundtruth_list = []
        self._detection_masks_list = []

    def add_single_ground_truth_image_info(self, image_id, groundtruth_dict):
        """Adds groundtruth for a single image to be used for evaluation.

        If the image has already been added, a warning is logged, and groundtruth is
        ignored.

        Args:
            image_id: A unique string/integer identifier for the image.
            groundtruth_dict: A dictionary containing
                :InputDataFields.groundtruth_boxes: float32 numpy array of shape
                    [num_boxes, 4] containing `num_boxes` groundtruth boxes of the format
                    [ymin, xmin, ymax, xmax] in absolute image coordinates.
                :InputDataFields.groundtruth_classes: integer numpy array of shape
                    [num_boxes] containing 1-indexed groundtruth classes for the boxes.
                :InputDataFields.groundtruth_instance_masks: uint8 numpy array of shape
                    [num_boxes, image_height, image_width] containing groundtruth masks
                    corresponding to the boxes. The elements of the array must be in
                    {0, 1}.
        """
        if image_id in self._image_id_to_mask_shape_map:
            print(
                'Ignoring ground truth with image id %s since it was '
                'previously added', image_id)
            return

        groundtruth_instance_masks = groundtruth_dict[
            standard_fields.InputDataFields.groundtruth_instance_masks]
        _check_mask_type_and_value(
            standard_fields.InputDataFields.groundtruth_instance_masks,
            groundtruth_instance_masks)
        self._groundtruth_list.extend(
            coco_tools.ExportSingleImageGroundtruthToCoco(
                image_id=image_id,
                next_annotation_id=self._annotation_id,
                category_id_set=self._category_id_set,
                groundtruth_boxes=groundtruth_dict[
                    standard_fields.InputDataFields.groundtruth_boxes],
                groundtruth_classes=groundtruth_dict[
                    standard_fields.InputDataFields.groundtruth_classes],
                groundtruth_masks=groundtruth_instance_masks,
                groundtruth_is_crowd=groundtruth_dict.get(
                    standard_fields.InputDataFields.groundtruth_is_crowd,
                    None)))
        self._annotation_id += groundtruth_dict[
            standard_fields.InputDataFields.groundtruth_boxes].shape[0]
        self._image_id_to_mask_shape_map[image_id] = groundtruth_dict[
            standard_fields.InputDataFields.groundtruth_instance_masks].shape

    def add_single_detected_image_info(self, image_id, detections_dict):
        """Adds detections for a single image to be used for evaluation.

        If a detection has already been added for this image id, a warning is
        logged, and the detection is skipped.

        Args:
            image_id: A unique string/integer identifier for the image.
            detections_dict: A dictionary containing -
                DetectionResultFields.detection_scores: float32 numpy array of shape
                [num_boxes] containing detection scores for the boxes.
                DetectionResultFields.detection_classes: integer numpy array of shape
                [num_boxes] containing 1-indexed detection classes for the boxes.
                DetectionResultFields.detection_masks: optional uint8 numpy array of
                shape [num_boxes, image_height, image_width] containing instance
                masks corresponding to the boxes. The elements of the array must be
                in {0, 1}.

        Raises:
            ValueError: If groundtruth for the image_id is not available or if
                spatial shapes of groundtruth_instance_masks and detection_masks are
                incompatible.
        """
        if image_id not in self._image_id_to_mask_shape_map:
            raise ValueError(
                'Missing groundtruth for image id: {}'.format(image_id))

        if image_id in self._image_ids_with_detections:
            print(
                'Ignoring detection with image id %s since it was '
                'previously added', image_id)
            return

        groundtruth_masks_shape = self._image_id_to_mask_shape_map[image_id]
        detection_masks = detections_dict[
            standard_fields.DetectionResultFields.detection_masks]
        if (len(detection_masks) and groundtruth_masks_shape[0] != 0
                and groundtruth_masks_shape[1:] != detection_masks.shape[1:]):
            raise ValueError(
                'Spatial shape of groundtruth masks and detection masks '
                'are incompatible: {} vs {}'.format(groundtruth_masks_shape,
                                                    detection_masks.shape))
        _check_mask_type_and_value(
            standard_fields.DetectionResultFields.detection_masks,
            detection_masks)
        self._detection_masks_list.extend(
            coco_tools.ExportSingleImageDetectionMasksToCoco(
                image_id=image_id,
                category_id_set=self._category_id_set,
                detection_masks=detection_masks,
                detection_scores=detections_dict[
                    standard_fields.DetectionResultFields.detection_scores],
                detection_classes=detections_dict[
                    standard_fields.DetectionResultFields.detection_classes]))
        self._image_ids_with_detections.update([image_id])

    def _evaluate(self):
        """Evaluates the detection masks and returns a dictionary of coco metrics.

        Returns:
            A dictionary holding -

            1. summary_metrics:
                'DetectionMasks_Precision/mAP': mean average precision over classes
                    averaged over IOU thresholds ranging from .5 to .95 with .05 increments.
                'DetectionMasks_Precision/mAP@.50IOU': mean average precision at 50% IOU.
                'DetectionMasks_Precision/mAP@.75IOU': mean average precision at 75% IOU.
                'DetectionMasks_Precision/mAP (small)': mean average precision for small
                    objects (area < 32^2 pixels).
                'DetectionMasks_Precision/mAP (medium)': mean average precision for medium
                    sized objects (32^2 pixels < area < 96^2 pixels).
                'DetectionMasks_Precision/mAP (large)': mean average precision for large
                    objects (96^2 pixels < area < 10000^2 pixels).
                'DetectionMasks_Recall/AR@1': average recall with 1 detection.
                'DetectionMasks_Recall/AR@10': average recall with 10 detections.
                'DetectionMasks_Recall/AR@100': average recall with 100 detections.
                'DetectionMasks_Recall/AR@100 (small)': average recall for small objects
                    with 100 detections.
                'DetectionMasks_Recall/AR@100 (medium)': average recall for medium objects
                    with 100 detections.
                'DetectionMasks_Recall/AR@100 (large)': average recall for large objects
                    with 100 detections.

            2. per_category_ap: if include_metrics_per_category is True, category
            specific results with keys of the form:
                'Precision mAP ByCategory/category' (without the supercategory part if
                no supercategories exist). For backward compatibility
                'PerformanceByCategory' is included in the output regardless of
                all_metrics_per_category.
        """
        groundtruth_dict = {
            'annotations':
            self._groundtruth_list,
            'images': [{
                'id': image_id,
                'height': shape[1],
                'width': shape[2]
            } for image_id, shape in six.iteritems(
                self._image_id_to_mask_shape_map)],
            'categories':
            self._categories
        }
        coco_wrapped_groundtruth = coco_tools.COCOWrapper(
            groundtruth_dict, detection_type='segmentation')
        coco_wrapped_detection_masks = coco_wrapped_groundtruth.LoadAnnotations(
            self._detection_masks_list)
        mask_evaluator = coco_tools.COCOEvalWrapper(
            coco_wrapped_groundtruth,
            coco_wrapped_detection_masks,
            agnostic_mode=False,
            iou_type='segm')
        mask_metrics, mask_per_category_ap = mask_evaluator.ComputeMetrics(
            include_metrics_per_category=self._include_metrics_per_category)
        mask_metrics.update(mask_per_category_ap)
        mask_metrics = {
            'DetectionMasks_' + key: value
            for key, value in six.iteritems(mask_metrics)
        }
        return mask_metrics

    def _evaluate_impl(self, prediction_dict, groundtruth_dict):
        """Evaluate with prediction and groundtruth dict

        Args:
            detections_dict: A dictionary containing -
                :DetectionResultFields.detection_scores: float32 numpy array of shape
                    [num_boxes] containing detection scores for the boxes.
                :DetectionResultFields.detection_classes: integer numpy array of shape
                    [num_boxes] containing 1-indexed detection classes for the boxes.
                :DetectionResultFields.detection_masks: mask rle code or optional uint8
                    numpy array of shape [num_boxes, image_height, image_width] containing
                    instance masks corresponding to the boxes. The elements of the array
                    must be in {0, 1}.
                :img_metas: List of length number of test images,
                    dict of image meta info, containing filename, ori_img_shape, and so on.
            groundtruth_dict: A dictionary containing
                :InputDataFields.groundtruth_boxes: float32 numpy array of shape
                    [num_boxes, 4] containing `num_boxes` groundtruth boxes of the format
                    [ymin, xmin, ymax, xmax] in absolute image coordinates.
                :InputDataFields.groundtruth_classes: integer numpy array of shape
                    [num_boxes] containing 1-indexed groundtruth classes for the boxes.
                :InputDataFields.groundtruth_instance_masks: mask rle code or uint8
                    numpy array of shape [num_boxes, image_height, image_width] containing
                    groundtruth masks corresponding to the boxes. The elements of the array
                    must be in {0, 1}.
                :groundtruth_is_crowd: integer numpy array of
                    shape [num_boxes] containing iscrowd flag for groundtruth boxes.
        """
        num_images_det = len(
            prediction_dict[DetectionResultFields.detection_boxes])
        num_images_gt = len(
            groundtruth_dict[InputDataFields.groundtruth_boxes])
        assert num_images_det == num_images_gt, 'detection and groundtruth number is not the same'

        groundtruth_is_crowd_list = groundtruth_dict.get(
            'groundtruth_is_crowd', None)
        for idx, (detection_masks, detection_classes, detection_scores, img_metas, gt_boxes, gt_masks, gt_classes) in \
            enumerate(zip(prediction_dict['detection_masks'], prediction_dict['detection_classes'],
                          prediction_dict['detection_scores'], prediction_dict['img_metas'],
                          groundtruth_dict['groundtruth_boxes'], groundtruth_dict['groundtruth_instance_masks'],
                          groundtruth_dict['groundtruth_classes'])):

            height, width = img_metas[
                'ori_img_shape'][:2]  # height, width, channel
            image_id = img_metas['filename']
            # tensor caculated on other devices is not on device:0, so transfer them
            if isinstance(detection_masks, torch.Tensor):
                detection_masks = detection_masks.cpu().numpy()
                detection_scores = detection_scores.cpu().numpy()
                detection_classes = detection_classes.cpu().numpy().astype(
                    np.int32)

            if groundtruth_is_crowd_list is None:
                groundtruth_is_crowd = None
            else:
                groundtruth_is_crowd = groundtruth_is_crowd_list[idx]

            if len(gt_masks) == 0:
                gt_masks = np.array([], dtype=np.uint8).reshape(
                    (0, height, width))
            else:
                gt_masks = np.array([
                    self._ann_to_mask(mask, height, width) for mask in gt_masks
                ],
                                    dtype=np.uint8)
            groundtruth_dict = {
                'groundtruth_boxes': gt_boxes,
                'groundtruth_instance_masks': gt_masks,
                'groundtruth_classes': gt_classes,
                'groundtruth_is_crowd': groundtruth_is_crowd,
            }
            self.add_single_ground_truth_image_info(image_id, groundtruth_dict)

            detection_masks = np.array([
                self._ann_to_mask(mask, height, width)
                for mask in detection_masks
            ],
                                       dtype=np.uint8)
            # add detection info
            detection_dict = {
                'detection_masks': detection_masks,
                'detection_scores': detection_scores,
                'detection_classes': detection_classes,
            }
            self.add_single_detected_image_info(image_id, detection_dict)

        eval_dict = self._evaluate()
        self.clear()
        return eval_dict

    def _ann_to_mask(self, segmentation, height, width):
        from xtcocotools import mask as maskUtils
        segm = segmentation
        h = height
        w = width

        if type(segm) == list:
            # polygon -- a single object might consist of multiple parts
            # we merge all parts into one mask rle code
            rles = maskUtils.frPyObjects(segm, h, w)
            rle = maskUtils.merge(rles)
        elif type(segm['counts']) == list:
            # uncompressed RLE
            rle = maskUtils.frPyObjects(segm, h, w)
        else:
            # rle
            rle = segm

        m = maskUtils.decode(rle)
        return m


@EVALUATORS.register_module
class CoCoPoseTopDownEvaluator(Evaluator):
    """Class to evaluate COCO keypoint topdown metrics.
    """

    def __init__(self, dataset_name=None, metric_names=['AP'], **kwargs):
        super().__init__(dataset_name, metric_names)

        self.vis_thr = kwargs.get('vis_thr', 0.2)
        self.oks_thr = kwargs.get('oks_thr', 0.9)
        self.use_nms = kwargs.get('use_nms', True)
        self.soft_nms = kwargs.get('soft_nms', False)

        metric = self.metric_names
        metrics = metric if isinstance(metric, list) else [metric]
        allowed_metrics = ['AP']
        for metric in metrics:
            if metric not in allowed_metrics:
                raise KeyError(f'metric {metric} is not supported')

    def _evaluate_impl(self, prediction_dict, groundtruth_dict, **kwargs):
        """
        Args:
            prediction_dict
                :preds (np.ndarray[N,K,3]): The first two dimensions are
                    coordinates, score is the third dimension of the array.
                :boxes (np.ndarray[N,6]): [center[0], center[1], scale[0]
                    , scale[1],area, score]
                :image_ids (list[int]): For example, [1, 2, ...]
                :heatmap (np.ndarray[N, K, H, W]): model output heatmap
                :bbox_id (list(int)): bbox_id

            groundtruth_dict (only support coco style)
                :images (list[dict]): dict keywords: 'file_name', 'height', 'width', 'id' and others
                :annotaions (list[dict]): dict keywords: 'num_keypoints', 'keypoints', 'image_id' and others
                :categories (list[dict]): list of category info

        """
        num_joints = kwargs['num_joints']
        class2id = kwargs['class2id']
        sigmas = kwargs.get('sigmas', None)

        image_ids = prediction_dict['image_ids']
        preds = prediction_dict['preds']
        boxes = prediction_dict['boxes']
        bbox_ids = prediction_dict['bbox_ids']

        kpts = defaultdict(list)
        for i, image_id in enumerate(image_ids):
            kpts[image_id].append({
                'keypoints': preds[i],
                'center': boxes[i][0:2],
                'scale': boxes[i][2:4],
                'area': boxes[i][4],
                'score': boxes[i][5],
                'image_id': image_id,
                'bbox_id': bbox_ids[i]
            })
        kpts = self._sort_and_unique_bboxes(kpts)

        # rescoring and oks nms
        valid_kpts = []
        for image_id in kpts.keys():
            img_kpts = kpts[image_id]
            for n_p in img_kpts:
                box_score = n_p['score']
                kpt_score = 0
                valid_num = 0
                for n_jt in range(0, num_joints):
                    t_s = n_p['keypoints'][n_jt][2]
                    if t_s > self.vis_thr:
                        kpt_score = kpt_score + t_s
                        valid_num = valid_num + 1
                if valid_num != 0:
                    kpt_score = kpt_score / valid_num
                # rescoring
                n_p['score'] = kpt_score * box_score

            if self.use_nms:
                nms = soft_oks_nms if self.soft_nms else oks_nms
                keep = nms(img_kpts, self.oks_thr, sigmas=sigmas)
                valid_kpts.append([img_kpts[_keep] for _keep in keep])
            else:
                valid_kpts.append(img_kpts)

        data_pack = [{
            'cat_id': cat_id,
            'ann_type': 'keypoints',
            'keypoints': valid_kpts
        } for cls, cat_id in class2id.items() if not cls == '__background__']

        results = self._coco_keypoint_results_one_category_kernel(
            data_pack[0], num_joints)
        info_str = self._do_python_keypoint_eval(results, groundtruth_dict,
                                                 sigmas)
        name_value = OrderedDict(info_str)

        return name_value

    def _sort_and_unique_bboxes(self, kpts, key='bbox_id'):
        """sort kpts and remove the repeated ones."""
        for img_id, persons in kpts.items():
            num = len(persons)
            kpts[img_id] = sorted(kpts[img_id], key=lambda x: x[key])
            for i in range(num - 1, 0, -1):
                if kpts[img_id][i][key] == kpts[img_id][i - 1][key]:
                    del kpts[img_id][i]

        return kpts

    def _coco_keypoint_results_one_category_kernel(self, data_pack,
                                                   num_joints):
        """Get coco keypoint results."""
        cat_id = data_pack['cat_id']
        keypoints = data_pack['keypoints']
        cat_results = []

        for img_kpts in keypoints:
            if len(img_kpts) == 0:
                continue

            _key_points = np.array(
                [img_kpt['keypoints'] for img_kpt in img_kpts])
            key_points = _key_points.reshape(-1, num_joints * 3)

            result = [{
                'image_id': img_kpt['image_id'],
                'category_id': cat_id,
                'keypoints': key_point.tolist(),
                'score': float(img_kpt['score']),
                'center': img_kpt['center'].tolist(),
                'scale': img_kpt['scale'].tolist()
            } for img_kpt, key_point in zip(img_kpts, key_points)]

            cat_results.extend(result)

        return cat_results

    def _do_python_keypoint_eval(self, results, groundtruth, sigmas=None):
        """Keypoint evaluation using COCOAPI."""

        with tempfile.TemporaryDirectory() as tmp_dir:
            groundtruth_file = os.path.join(tmp_dir,
                                            'groundtruth_keypoints.json')
            with open(groundtruth_file, 'w') as f:
                json.dump(groundtruth, f, sort_keys=True, indent=4)
            coco = COCO(groundtruth_file)

            res_file = os.path.join(tmp_dir, 'result_keypoints.json')
            with open(res_file, 'w') as f:
                json.dump(results, f, sort_keys=True, indent=4, cls=MyEncoder)
            coco_det = coco.loadRes(res_file)

        coco_eval = COCOeval(coco, coco_det, 'keypoints', sigmas=sigmas)
        coco_eval.params.useSegm = None
        coco_eval.evaluate()
        coco_eval.accumulate()
        coco_eval.summarize()

        stats_names = [
            'AP', 'AP .5', 'AP .75', 'AP (M)', 'AP (L)', 'AR', 'AR .5',
            'AR .75', 'AR (M)', 'AR (L)'
        ]

        info_str = list(zip(stats_names, coco_eval.stats))

        return info_str


@EVALUATORS.register_module
class CocoPanopticEvaluator(Evaluator):
    """Class to evaluate COCO panoptic metrics.
    """

    def __init__(self,
                 dataset_name=None,
                 metric_names=['PQ'],
                 classes=None,
                 file_client_args=dict(backend='disk'),
                 **kwargs):
        super().__init__(dataset_name, metric_names)
        self.CLASSES = classes
        self.file_client = mmcv.FileClient(**file_client_args)

    def evaluate(self,
                 gt_json,
                 gt_folder,
                 pred_json,
                 pred_folder,
                 categories,
                 nproc=32,
                 classwise=False,
                 **kwargs):

        # match the gt_anns and pred_anns in the same image
        matched_annotations_list = []
        for gt_ann in gt_json:
            img_id = gt_ann['image_id']
            if img_id not in pred_json.keys():
                continue
                # raise Exception('no prediction for the image'
                #                 ' with id: {}'.format(img_id))
            matched_annotations_list.append((gt_ann, pred_json[img_id]))

        pq_stat = pq_compute_multi_core(
            matched_annotations_list,
            gt_folder,
            pred_folder,
            categories,
            self.file_client,
            nproc=nproc)

        metrics = [('All', None), ('Things', True), ('Stuff', False)]
        pq_results = {}

        for name, isthing in metrics:
            pq_results[name], classwise_results = pq_stat.pq_average(
                categories, isthing=isthing)
            if name == 'All':
                pq_results['classwise'] = classwise_results
        classwise_results = None
        if classwise:
            classwise_results = {
                k: v
                for k, v in zip(self.CLASSES, pq_results['classwise'].values())
            }
        results = self.parse_pq_results(pq_results)
        return results

    def parse_pq_results(self, pq_results):
        """Parse the Panoptic Quality results."""
        result = dict()
        result['PQ'] = 100 * pq_results['All']['pq']
        result['SQ'] = 100 * pq_results['All']['sq']
        result['RQ'] = 100 * pq_results['All']['rq']
        result['PQ_th'] = 100 * pq_results['Things']['pq']
        result['SQ_th'] = 100 * pq_results['Things']['sq']
        result['RQ_th'] = 100 * pq_results['Things']['rq']
        result['PQ_st'] = 100 * pq_results['Stuff']['pq']
        result['SQ_st'] = 100 * pq_results['Stuff']['sq']
        result['RQ_st'] = 100 * pq_results['Stuff']['rq']
        return result

    def _evaluate_impl(self, predictions, labels, **kwargs):
        pass


def pq_compute_single_core(proc_id,
                           annotation_set,
                           gt_folder,
                           pred_folder,
                           categories,
                           file_client=None,
                           print_log=False):
    """The single core function to evaluate the metric of Panoptic
    Segmentation.

    Same as the function with the same name in `panopticapi`. Only the function
    to load the images is changed to use the file client.

    Args:
        proc_id (int): The id of the mini process.
        gt_folder (str): The path of the ground truth images.
        pred_folder (str): The path of the prediction images.
        categories (str): The categories of the dataset.
        file_client (object): The file client of the dataset. If None,
            the backend will be set to `disk`.
        print_log (bool): Whether to print the log. Defaults to False.
    """
    if PQStat is None:
        raise RuntimeError(
            'panopticapi is not installed, please install it by: '
            'pip install git+https://github.com/cocodataset/'
            'panopticapi.git.')

    if file_client is None:
        file_client_args = dict(backend='disk')
        file_client = mmcv.FileClient(**file_client_args)

    pq_stat = PQStat()

    idx = 0
    for gt_ann, pred_ann in annotation_set:
        if print_log and idx % 100 == 0:
            print('Core: {}, {} from {} images processed'.format(
                proc_id, idx, len(annotation_set)))
        idx += 1
        # The gt images can be on the local disk or `ceph`, so we use
        # file_client here.
        img_bytes = file_client.get(
            os.path.join(gt_folder, gt_ann['file_name']))
        pan_gt = mmcv.imfrombytes(img_bytes, flag='color', channel_order='rgb')
        pan_gt = rgb2id(pan_gt)

        # The predictions can only be on the local dist now.
        pan_pred = mmcv.imread(
            os.path.join(pred_folder, pred_ann['file_name']),
            flag='color',
            channel_order='rgb')
        pan_pred = rgb2id(pan_pred)

        gt_segms = {el['id']: el for el in gt_ann['segments_info']}
        pred_segms = {el['id']: el for el in pred_ann['segments_info']}

        # predicted segments area calculation + prediction sanity checks
        pred_labels_set = set(el['id'] for el in pred_ann['segments_info'])
        labels, labels_cnt = np.unique(pan_pred, return_counts=True)
        for label, label_cnt in zip(labels, labels_cnt):
            if label not in pred_segms:
                if label == VOID:
                    continue
                raise KeyError(
                    'In the image with ID {} segment with ID {} is '
                    'presented in PNG and not presented in JSON.'.format(
                        gt_ann['image_id'], label))
            pred_segms[label]['area'] = label_cnt
            pred_labels_set.remove(label)
            if pred_segms[label]['category_id'] not in categories:
                raise KeyError(
                    'In the image with ID {} segment with ID {} has '
                    'unknown category_id {}.'.format(
                        gt_ann['image_id'], label,
                        pred_segms[label]['category_id']))
        if len(pred_labels_set) != 0:
            raise KeyError(
                'In the image with ID {} the following segment IDs {} '
                'are presented in JSON and not presented in PNG.'.format(
                    gt_ann['image_id'], list(pred_labels_set)))

        # confusion matrix calculation
        pan_gt_pred = pan_gt.astype(np.uint64) * OFFSET + pan_pred.astype(
            np.uint64)
        gt_pred_map = {}
        labels, labels_cnt = np.unique(pan_gt_pred, return_counts=True)
        for label, intersection in zip(labels, labels_cnt):
            gt_id = label // OFFSET
            pred_id = label % OFFSET
            gt_pred_map[(gt_id, pred_id)] = intersection

        # count all matched pairs
        gt_matched = set()
        pred_matched = set()
        for label_tuple, intersection in gt_pred_map.items():
            gt_label, pred_label = label_tuple
            if gt_label not in gt_segms:
                continue
            if pred_label not in pred_segms:
                continue
            if gt_segms[gt_label]['iscrowd'] == 1:
                continue
            if gt_segms[gt_label]['category_id'] != pred_segms[pred_label][
                    'category_id']:
                continue

            union = pred_segms[pred_label]['area'] + gt_segms[gt_label][
                'area'] - intersection - gt_pred_map.get((VOID, pred_label), 0)
            iou = intersection / union
            if iou > 0.5:
                pq_stat[gt_segms[gt_label]['category_id']].tp += 1
                pq_stat[gt_segms[gt_label]['category_id']].iou += iou
                gt_matched.add(gt_label)
                pred_matched.add(pred_label)

        # count false positives
        crowd_labels_dict = {}
        for gt_label, gt_info in gt_segms.items():
            if gt_label in gt_matched:
                continue
            # crowd segments are ignored
            if gt_info['iscrowd'] == 1:
                crowd_labels_dict[gt_info['category_id']] = gt_label
                continue
            pq_stat[gt_info['category_id']].fn += 1

        # count false positives
        for pred_label, pred_info in pred_segms.items():
            if pred_label in pred_matched:
                continue
            # intersection of the segment with VOID
            intersection = gt_pred_map.get((VOID, pred_label), 0)
            # plus intersection with corresponding CROWD region if it exists
            if pred_info['category_id'] in crowd_labels_dict:
                intersection += gt_pred_map.get(
                    (crowd_labels_dict[pred_info['category_id']], pred_label),
                    0)
            # predicted segment is ignored if more than half of
            # the segment correspond to VOID and CROWD regions
            if intersection / pred_info['area'] > 0.5:
                continue
            pq_stat[pred_info['category_id']].fp += 1

    if print_log:
        print('Core: {}, all {} images processed'.format(
            proc_id, len(annotation_set)))
    return pq_stat


def pq_compute_multi_core(matched_annotations_list,
                          gt_folder,
                          pred_folder,
                          categories,
                          file_client=None,
                          nproc=32):
    """Evaluate the metrics of Panoptic Segmentation with multithreading.

    Same as the function with the same name in `panopticapi`.

    Args:
        matched_annotations_list (list): The matched annotation list. Each
            element is a tuple of annotations of the same image with the
            format (gt_anns, pred_anns).
        gt_folder (str): The path of the ground truth images.
        pred_folder (str): The path of the prediction images.
        categories (str): The categories of the dataset.
        file_client (object): The file client of the dataset. If None,
            the backend will be set to `disk`.
        nproc (int): Number of processes for panoptic quality computing.
            Defaults to 32. When `nproc` exceeds the number of cpu cores,
            the number of cpu cores is used.
    """
    if PQStat is None:
        raise RuntimeError(
            'panopticapi is not installed, please install it by: '
            'pip install git+https://github.com/cocodataset/'
            'panopticapi.git.')

    if file_client is None:
        file_client_args = dict(backend='disk')
        file_client = mmcv.FileClient(**file_client_args)

    cpu_num = min(nproc, multiprocessing.cpu_count())

    annotations_split = np.array_split(matched_annotations_list, cpu_num)
    print('Number of cores: {}, images per core: {}'.format(
        cpu_num, len(annotations_split[0])))
    workers = multiprocessing.Pool(processes=cpu_num)
    processes = []
    for proc_id, annotation_set in enumerate(annotations_split):
        p = workers.apply_async(pq_compute_single_core,
                                (proc_id, annotation_set, gt_folder,
                                 pred_folder, categories, file_client))
        processes.append(p)

    # Close the process pool, otherwise it will lead to memory
    # leaking problems.
    workers.close()
    workers.join()

    pq_stat = PQStat()
    for p in processes:
        pq_stat += p.get()

    return pq_stat


METRICS.register_default_best_metric(CocoDetectionEvaluator,
                                     'DetectionBoxes_Precision/mAP', 'max')
METRICS.register_default_best_metric(CocoMaskEvaluator,
                                     'DetectionMasks_Precision/mAP', 'max')
METRICS.register_default_best_metric(CoCoPoseTopDownEvaluator, 'AP', 'max')
METRICS.register_default_best_metric(CocoPanopticEvaluator, 'PQ', 'max')



_base_ = '../../base.py'

# model settings
model = dict(
    type='MOCO',
    pretrained=False,
    train_preprocess=['gaussianBlur'],
    queue_len=65536,
    feat_dim=128,
    momentum=0.999,
    backbone=dict(
        type='ResNet',
        depth=50,
        in_channels=3,
        out_indices=[4],  # 0: conv-1, x: stage-x
        norm_cfg=dict(type='BN')),
    neck=dict(
        type='NonLinearNeckV1',
        in_channels=2048,
        hid_channels=2048,
        out_channels=128,
        with_avg_pool=True),
    head=dict(type='ContrastiveHead', temperature=0.2))

# dataset settings
data_train_list = 'imagenet_raw/meta/train.txt'
data_train_root = 'imagenet_raw/'
dataset_type = 'MultiViewDataset'
img_norm_cfg = dict(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])
train_pipeline = [
    dict(type='RandomResizedCrop', size=224, scale=(0.2, 1.)),
    dict(
        type='RandomAppliedTrans',
        transforms=[
            dict(
                type='ColorJitter',
                brightness=0.4,
                contrast=0.4,
                saturation=0.4,
                hue=0.4)
        ],
        p=0.8),
    dict(type='RandomGrayscale', p=0.2),
    # dict(
    # type='RandomAppliedTrans',
    # transforms=[
    #     dict(
    #         type='GaussianBlur',
    #         kernel_size=23,
    #         sigma = (0.1, 2.0)
    #         )
    # ],
    # p=0.5),
    dict(type='RandomHorizontalFlip'),
    dict(type='ToTensor'),
    dict(type='Normalize', **img_norm_cfg),
    dict(type='Collect', keys=['img'])
]
data = dict(
    imgs_per_gpu=32,  # total 32*8=256
    workers_per_gpu=4,
    drop_last=True,
    train=dict(
        type=dataset_type,
        data_source=dict(
            type='SSLSourceImageList',
            list_file=data_train_list,
            root=data_train_root),
        num_views=[1, 1],
        pipelines=[train_pipeline, train_pipeline]))
# optimizer
optimizer = dict(type='SGD', lr=0.03, weight_decay=0.0001, momentum=0.9)
# learning policy
lr_config = dict(policy='step', step=[120, 160])
# lr_config = dict(policy='CosineAnnealing', min_lr=0.)
checkpoint_config = dict(interval=10)
# runtime settings
total_epochs = 200



# Copyright (c) Alibaba, Inc. and its affiliates.
import torch
import torch.nn as nn
from mmcv.cnn import normal_init

from .. import builder
from ..registry import HEADS


@HEADS.register_module
class LatentPredictHead(nn.Module):
    '''Head for contrastive learning.
    '''

    def __init__(self, predictor, size_average=True):
        super(LatentPredictHead, self).__init__()
        self.predictor = builder.build_neck(predictor)
        self.size_average = size_average

    def init_weights(self, init_linear='normal'):
        self.predictor.init_weights(init_linear=init_linear)

    def forward(self, input, target):
        '''
        Args:
            input (Tensor): NxC input features.
            target (Tensor): NxC target features.
        '''
        pred = self.predictor([input])[0]
        pred_norm = nn.functional.normalize(pred, dim=1)
        target_norm = nn.functional.normalize(target, dim=1)
        loss = -2 * (pred_norm * target_norm).sum()
        if self.size_average:
            loss /= input.size(0)
        return dict(loss=loss)


@HEADS.register_module
class LatentClsHead(nn.Module):
    '''Head for contrastive learning.
    '''

    def __init__(self, predictor):
        super(LatentClsHead, self).__init__()
        self.predictor = nn.Linear(predictor.in_channels,
                                   predictor.num_classes)
        self.criterion = nn.CrossEntropyLoss()

    def init_weights(self, init_linear='normal'):
        normal_init(self.predictor, std=0.01)

    def forward(self, input, target):
        '''
        Args:
            input (Tensor): NxC input features.
            target (Tensor): NxC target features.
        '''
        pred = self.predictor(input)
        with torch.no_grad():
            label = torch.argmax(self.predictor(target), dim=1).detach()
        loss = self.criterion(pred, label)
        return dict(loss=loss)



import torch
import torch.nn.functional as F
from scipy.optimize import linear_sum_assignment
from torch import nn
from torch.cuda.amp import autocast

from .point_rend import point_sample


def batch_dice_loss(inputs: torch.Tensor, targets: torch.Tensor):
    """
    Compute the DICE loss, similar to generalized IOU for masks
    Args:
        inputs: A float tensor of arbitrary shape.
                The predictions for each example.
        targets: A float tensor with the same shape as inputs. Stores the binary
                 classification label for each element in inputs
                (0 for the negative class and 1 for the positive class).
    """
    inputs = inputs.sigmoid()
    inputs = inputs.flatten(1)
    numerator = 2 * torch.einsum('nc,mc->nm', inputs, targets)
    denominator = inputs.sum(-1)[:, None] + targets.sum(-1)[None, :]
    loss = 1 - (numerator + 1) / (denominator + 1)
    return loss


def batch_sigmoid_ce_loss(inputs: torch.Tensor, targets: torch.Tensor):
    """
    Args:
        inputs: A float tensor of arbitrary shape.
                The predictions for each example.
        targets: A float tensor with the same shape as inputs. Stores the binary
                 classification label for each element in inputs
                (0 for the negative class and 1 for the positive class).
    Returns:
        Loss tensor
    """
    hw = inputs.shape[1]

    pos = F.binary_cross_entropy_with_logits(
        inputs, torch.ones_like(inputs), reduction='none')
    neg = F.binary_cross_entropy_with_logits(
        inputs, torch.zeros_like(inputs), reduction='none')

    loss = torch.einsum('nc,mc->nm', pos, targets) + torch.einsum(
        'nc,mc->nm', neg, (1 - targets))
    return loss / hw


batch_sigmoid_ce_loss_jit = torch.jit.script(
    batch_sigmoid_ce_loss)  # type: torch.jit.ScriptModule


# Modified from https://github.com/facebookresearch/detr/blob/master/models/matcher.py
class MaskHungarianMatcher(nn.Module):
    """This class computes an assignment between the targets and the predictions of the network for segmentation

    For efficiency reasons, the targets don't include the no_object. Because of this, in general,
    there are more predictions than targets. In this case, we do a 1-to-1 matching of the best predictions,
    while the others are un-matched (and thus treated as non-objects).
    """

    def __init__(self,
                 cost_class: float = 1,
                 cost_mask: float = 1,
                 cost_dice: float = 1,
                 num_points: int = 0):
        """Creates the matcher

        Params:
            cost_class: This is the relative weight of the classification error in the matching cost
            cost_mask: This is the relative weight of the focal loss of the binary mask in the matching cost
            cost_dice: This is the relative weight of the dice loss of the binary mask in the matching cost
        """
        super().__init__()
        self.cost_class = cost_class
        self.cost_mask = cost_mask
        self.cost_dice = cost_dice

        assert cost_class != 0 or cost_mask != 0 or cost_dice != 0, 'all costs cant be 0'

        self.num_points = num_points

    @torch.no_grad()
    def memory_efficient_forward(self, outputs, targets):
        """More memory-friendly matching"""
        bs, num_queries = outputs['pred_logits'].shape[:2]

        indices = []
        # Iterate through batch size
        for b in range(bs):

            out_prob = outputs['pred_logits'][b].softmax(
                -1)  # [num_queries, num_classes]
            tgt_ids = targets[b]['labels']

            # Compute the classification cost. Contrary to the loss, we don't use the NLL,
            # but approximate it in 1 - proba[target class].
            # The 1 is a constant that doesn't change the matching, it can be ommitted.
            cost_class = -out_prob[:, tgt_ids]

            out_mask = outputs['pred_masks'][
                b]  # [num_queries, H_pred, W_pred]
            # gt masks are already padded when preparing target
            tgt_mask = targets[b]['masks'].to(out_mask)

            out_mask = out_mask[:, None]
            tgt_mask = tgt_mask[:, None]
            # all masks share the same set of points for efficient matching!
            point_coords = torch.rand(
                1, self.num_points, 2, device=out_mask.device)
            # get gt labels
            tgt_mask = point_sample(
                tgt_mask,
                point_coords.repeat(tgt_mask.shape[0], 1, 1),
                align_corners=False,
            ).squeeze(1)
            out_mask = point_sample(
                out_mask,
                point_coords.repeat(out_mask.shape[0], 1, 1),
                align_corners=False,
            ).squeeze(1)
            with autocast(enabled=False):
                out_mask = out_mask.float()
                tgt_mask = tgt_mask.float()
                # Compute the focal loss between masks
                # jit not match some torch versions
                # cost_mask = batch_sigmoid_ce_loss_jit(out_mask, tgt_mask)
                cost_mask = batch_sigmoid_ce_loss(out_mask, tgt_mask)

                # Compute the dice loss betwen masks
                # cost_dice = batch_dice_loss_jit(out_mask, tgt_mask)
                cost_dice = batch_dice_loss(out_mask, tgt_mask)

            # Final cost matrix
            C = (
                self.cost_mask * cost_mask + self.cost_class * cost_class +
                self.cost_dice * cost_dice)
            C = C.reshape(num_queries, -1).cpu()
            indices.append(linear_sum_assignment(C))

        return [(torch.as_tensor(i, dtype=torch.int64),
                 torch.as_tensor(j, dtype=torch.int64)) for i, j in indices]

    @torch.no_grad()
    def forward(self, outputs, targets):
        """Performs the matching

        Params:
            outputs: This is a dict that contains at least these entries:
                 "pred_logits": Tensor of dim [batch_size, num_queries, num_classes] with the classification logits
                 "pred_masks": Tensor of dim [batch_size, num_queries, H_pred, W_pred] with the predicted masks

            targets: This is a list of targets (len(targets) = batch_size), where each target is a dict containing:
                 "labels": Tensor of dim [num_target_boxes] (where num_target_boxes is the number of ground-truth
                           objects in the target) containing the class labels
                 "masks": Tensor of dim [num_target_boxes, H_gt, W_gt] containing the target masks

        Returns:
            A list of size batch_size, containing tuples of (index_i, index_j) where:
                - index_i is the indices of the selected predictions (in order)
                - index_j is the indices of the corresponding selected targets (in order)
            For each batch element, it holds:
                len(index_i) = len(index_j) = min(num_queries, num_target_boxes)
        """
        return self.memory_efficient_forward(outputs, targets)

    def __repr__(self, _repr_indent=4):
        head = 'Matcher ' + self.__class__.__name__
        body = [
            'cost_class: {}'.format(self.cost_class),
            'cost_mask: {}'.format(self.cost_mask),
            'cost_dice: {}'.format(self.cost_dice),
        ]
        lines = [head] + [' ' * _repr_indent + line for line in body]
        return '\n'.join(lines)



# Copyright (c) Alibaba, Inc. and its affiliates.
from PIL import Image

from easycv.datasets.registry import DATASETS
from easycv.datasets.shared.base import BaseDataset
from easycv.framework.errors import NotImplementedError


@DATASETS.register_module
class ClsOdpsDataset(BaseDataset):
    """Dataset for rotation prediction
    """

    def __init__(self,
                 data_source,
                 pipeline,
                 image_key='url_image',
                 label_key='label',
                 **kwargs):
        super(ClsOdpsDataset, self).__init__(data_source, pipeline)
        self.image_key = image_key
        self.label_key = label_key

    def __getitem__(self, idx):
        data_dict = self.data_source[idx]
        assert (self.image_key in data_dict.keys())
        assert (self.label_key in data_dict.keys())

        img = data_dict[self.image_key]
        label = int(data_dict[self.label_key])
        assert isinstance(img, Image.Image), \
            f'The output from the data source must be an Image, got: {type(img)}. \
            Please ensure that the list file does not contain labels.'

        img = self.pipeline(img)
        return dict(img=img, gt_labels=label)

    def evaluate(self, results, evaluators, logger=None):
        raise NotImplementedError



from .detection_pipeline import EasyCVDetectionPipeline
from .face_2d_keypoints_pipeline import Face2DKeypointsPipeline
from .hand_2d_keypoints_pipeline import Hand2DKeypointsPipeline
from .human_wholebody_keypoint_pipeline import HumanWholebodyKeypointsPipeline
from .image_panoptic_segmentation_pipeline import \
    ImagePanopticSegmentationEasyCVPipeline
from .segmentation_pipeline import EasyCVSegmentationPipeline

__all__ = [
    'EasyCVDetectionPipeline', 'EasyCVSegmentationPipeline',
    'Face2DKeypointsPipeline', 'HumanWholebodyKeypointsPipeline',
    'Hand2DKeypointsPipeline', 'ImagePanopticSegmentationEasyCVPipeline'
]



# Copyright (c) Alibaba, Inc. and its affiliates.
from mmcv.runner.hooks import HOOKS
from mmcv.runner.hooks.lr_updater import (CosineAnnealingLrUpdaterHook,
                                          annealing_cos)


@HOOKS.register_module(force=True)
class YOLOXLrUpdaterHook(CosineAnnealingLrUpdaterHook):
    """YOLOX learning rate scheme.

    There are two main differences between YOLOXLrUpdaterHook
    and CosineAnnealingLrUpdaterHook.

       1. When the current running epoch is greater than
           `max_epoch-last_epoch`, a fixed learning rate will be used
       2. The exp warmup scheme is different with LrUpdaterHook in MMCV

    Args:
        num_last_epochs (int): The number of epochs with a fixed learning rate
           before the end of the training.
    """

    def __init__(self, num_last_epochs, **kwargs):
        self.num_last_epochs = num_last_epochs
        super(YOLOXLrUpdaterHook, self).__init__(**kwargs)

    def get_warmup_lr(self, cur_iters):

        def _get_warmup_lr(cur_iters, regular_lr):
            # exp warmup scheme
            k = self.warmup_ratio * pow(
                (cur_iters + 1) / float(self.warmup_iters), 2)
            warmup_lr = [_lr * k for _lr in regular_lr]
            return warmup_lr

        if isinstance(self.base_lr, dict):
            lr_groups = {}
            for key, base_lr in self.base_lr.items():
                lr_groups[key] = _get_warmup_lr(cur_iters, base_lr)
            return lr_groups
        else:
            return _get_warmup_lr(cur_iters, self.base_lr)

    def get_lr(self, runner, base_lr):
        last_iter = len(runner.data_loader) * self.num_last_epochs

        if self.by_epoch:
            progress = runner.epoch
            max_progress = runner.max_epochs
        else:
            progress = runner.iter
            max_progress = runner.max_iters

        progress += 1

        if self.min_lr_ratio is not None:
            target_lr = base_lr * self.min_lr_ratio
        else:
            target_lr = self.min_lr

        if progress >= max_progress - last_iter:
            # fixed learning rate
            return target_lr
        else:
            assert max_progress > (self.warmup_iters + last_iter
                                   ), 'Please increase the train epoch'
            return annealing_cos(
                base_lr, target_lr, (progress - self.warmup_iters) /
                (max_progress - self.warmup_iters - last_iter))



# Copyright 2017 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Utilities for dealing with writing json strings.

json_utils wraps json.dump and json.dumps so that they can be used to safely
control the precision of floats when writing to json strings or files.
"""
import json
import sys
from json import encoder

import numpy as np

from easycv.framework.errors import ValueError

# python 3.5 and newer version does not have json.encoder.FLOAT_REPR
needs_class_hack = sys.version_info >= (3, 5)

encoder.FLOAT_REPR = float.__repr__
INFINITY = float('inf')


class MyEncoder(encoder.JSONEncoder):

    def default(self, o):
        """Implement this method in a subclass such that it returns
    a serializable object for ``o``, or calls the base implementation
    (to raise a ``TypeError``).

    For example, to support arbitrary iterators, you could
    implement default like this::

        def default(self, o):
            try:
                iterable = iter(o)
            except TypeError:
                pass
            else:
                return list(iterable)
            # Let the base class default method raise the TypeError
            return JSONEncoder.default(self, o)

    """
        vtype = type(o)
        if isinstance(o, bytes):
            return str(o, encoding='utf-8')
        elif isinstance(o, np.ndarray):
            return o.tolist()
        elif vtype in [np.int16, np.int32, np.int64, np.float32, np.float64]:
            return o.item()
        else:
            return encoder.JSONEncoder.default(self, o)

    def iterencode(self, o, _one_shot=False):
        """Encode the given object and yield each string
    representation as available.

    For example::

        for chunk in JSONEncoder().iterencode(bigobject):
            mysocket.write(chunk)

    """
        if self.check_circular:
            markers = {}
        else:
            markers = None
        if self.ensure_ascii:
            _encoder = encoder.encode_basestring_ascii
        else:
            _encoder = encoder.encode_basestring

        def floatstr(o,
                     allow_nan=self.allow_nan,
                     _repr=encoder.FLOAT_REPR,
                     _inf=INFINITY,
                     _neginf=-INFINITY):
            # Check for specials.  Note that this type of test is processor
            # and/or platform-specific, so do tests which don't depend on the
            # internals.

            if o != o:
                text = 'NaN'
            elif o == _inf:
                text = 'Infinity'
            elif o == _neginf:
                text = '-Infinity'
            else:
                return _repr(o)

            if not allow_nan:
                raise ValueError(
                    'Out of range float values are not JSON compliant: ' +
                    repr(o))

            return text

        if (_one_shot and encoder.c_make_encoder is not None
                and self.indent is None):
            _iterencode = encoder.c_make_encoder(markers, self.default,
                                                 _encoder, self.indent,
                                                 self.key_separator,
                                                 self.item_separator,
                                                 self.sort_keys, self.skipkeys,
                                                 self.allow_nan)
        else:
            _iterencode = encoder._make_iterencode(
                markers, self.default, _encoder, self.indent, floatstr,
                self.key_separator, self.item_separator, self.sort_keys,
                self.skipkeys, _one_shot)
        return _iterencode(o, 0)


def dump(obj, fid, float_digits=-1, **params):
    """Wrapper of json.dump that allows specifying the float precision used.

  Args:
    obj: The object to dump.
    fid: The file id to write to.
    float_digits: The number of digits of precision when writing floats out.
    **params: Additional parameters to pass to json.dumps.
  """
    original_encoder = encoder.FLOAT_REPR
    if float_digits >= 0:
        encoder.FLOAT_REPR = lambda o: format(o, '.%df' % float_digits)
    try:
        if needs_class_hack and 'cls' not in params:
            params['cls'] = MyEncoder
        json.dump(obj, fid, **params)
    finally:
        encoder.FLOAT_REPR = original_encoder


def dumps(obj, float_digits=-1, **params):
    """Wrapper of json.dumps that allows specifying the float precision used.

  Args:
    obj: The object to dump.
    float_digits: The number of digits of precision when writing floats out.
    **params: Additional parameters to pass to json.dumps.

  Returns:
    output: JSON string representation of obj.
  """
    original_encoder = encoder.FLOAT_REPR
    original_c_make_encoder = encoder.c_make_encoder
    if float_digits >= 0:
        encoder.FLOAT_REPR = lambda o: format(o, '.%df' % float_digits)
        encoder.c_make_encoder = None
    try:
        if needs_class_hack and 'cls' not in params:
            params['cls'] = MyEncoder
        output = json.dumps(obj, **params)

    finally:
        encoder.FLOAT_REPR = original_encoder
        encoder.c_make_encoder = original_c_make_encoder

    return output


def compat_dumps(data, float_digits=-1):
    """
  handle json dumps chinese and numpy data
  Args:
    data python data structure
    float_digits: The number of digits of precision when writing floats out.
  Return:
    json str, in python2 , the str is encoded with utf8
      in python3, the str is unicode type(python3 str)
  """
    result_str = dumps(
        data, float_digits=float_digits, cls=MyEncoder, ensure_ascii=False)
    return result_str


def PrettyParams(**params):
    """Returns parameters for use with Dump and Dumps to output pretty json.

  Example usage:
    ```json_str = json_utils.Dumps(obj, **json_utils.PrettyParams())```
    ```json_str = json_utils.Dumps(
                      obj, **json_utils.PrettyParams(allow_nans=False))```

  Args:
    **params: Additional params to pass to json.dump or json.dumps.

  Returns:
    params: Parameters that are compatible with json_utils.Dump and
      json_utils.Dumps.
  """
    params['float_digits'] = 4
    params['sort_keys'] = True
    params['indent'] = 2
    params['separators'] = (',', ': ')
    return params



_base_ = '../../base.py'

# model settings s m l x
model = dict(
    type='YOLOX',
    test_conf=0.01,
    nms_thre=0.65,
    backbone='RepVGGYOLOX',
    model_type='s',  # s m l x tiny nano
    head=dict(
        type='YOLOXHead',
        model_type='s',
        obj_loss_type='BCE',
        reg_loss_type='giou',
        num_classes=80,
        decode_in_inference=
        True  # set to False when test speed to ignore decode and nms
    ))

# s m l x
img_scale = (640, 640)
random_size = (14, 26)
scale_ratio = (0.1, 2)

CLASSES = [
    'person', 'bicycle', 'car', 'motorcycle', 'airplane', 'bus', 'train',
    'truck', 'boat', 'traffic light', 'fire hydrant', 'stop sign',
    'parking meter', 'bench', 'bird', 'cat', 'dog', 'horse', 'sheep', 'cow',
    'elephant', 'bear', 'zebra', 'giraffe', 'backpack', 'umbrella', 'handbag',
    'tie', 'suitcase', 'frisbee', 'skis', 'snowboard', 'sports ball', 'kite',
    'baseball bat', 'baseball glove', 'skateboard', 'surfboard',
    'tennis racket', 'bottle', 'wine glass', 'cup', 'fork', 'knife', 'spoon',
    'bowl', 'banana', 'apple', 'sandwich', 'orange', 'broccoli', 'carrot',
    'hot dog', 'pizza', 'donut', 'cake', 'chair', 'couch', 'potted plant',
    'bed', 'dining table', 'toilet', 'tv', 'laptop', 'mouse', 'remote',
    'keyboard', 'cell phone', 'microwave', 'oven', 'toaster', 'sink',
    'refrigerator', 'book', 'clock', 'vase', 'scissors', 'teddy bear',
    'hair drier', 'toothbrush'
]

# dataset settings
data_root = 'data/coco/'

img_norm_cfg = dict(
    mean=[123.675, 116.28, 103.53], std=[58.395, 57.12, 57.375], to_rgb=True)

train_pipeline = [
    dict(type='MMMosaic', img_scale=img_scale, pad_val=114.0),
    dict(
        type='MMRandomAffine',
        scaling_ratio_range=scale_ratio,
        border=(-img_scale[0] // 2, -img_scale[1] // 2)),
    dict(
        type='MMMixUp',  # s m x l; tiny nano will detele
        img_scale=img_scale,
        ratio_range=(0.8, 1.6),
        pad_val=114.0),
    dict(
        type='MMPhotoMetricDistortion',
        brightness_delta=32,
        contrast_range=(0.5, 1.5),
        saturation_range=(0.5, 1.5),
        hue_delta=18),
    dict(type='MMRandomFlip', flip_ratio=0.5),
    dict(type='MMResize', keep_ratio=True),
    dict(type='MMPad', pad_to_square=True, pad_val=(114.0, 114.0, 114.0)),
    dict(type='MMNormalize', **img_norm_cfg),
    dict(type='DefaultFormatBundle'),
    dict(type='Collect', keys=['img', 'gt_bboxes', 'gt_labels'])
]
test_pipeline = [
    dict(type='MMResize', img_scale=img_scale, keep_ratio=True),
    dict(type='MMPad', pad_to_square=True, pad_val=(114.0, 114.0, 114.0)),
    dict(type='MMNormalize', **img_norm_cfg),
    dict(type='DefaultFormatBundle'),
    dict(type='Collect', keys=['img'])
]

train_dataset = dict(
    type='DetImagesMixDataset',
    data_source=dict(
        type='DetSourceCoco',
        ann_file=data_root + 'annotations/instances_train2017.json',
        img_prefix=data_root + 'train2017/',
        pipeline=[
            dict(type='LoadImageFromFile', to_float32=True),
            dict(type='LoadAnnotations', with_bbox=True)
        ],
        classes=CLASSES,
        filter_empty_gt=True,
        iscrowd=False),
    pipeline=train_pipeline,
    dynamic_scale=img_scale)

val_dataset = dict(
    type='DetImagesMixDataset',
    imgs_per_gpu=2,
    data_source=dict(
        type='DetSourceCoco',
        ann_file=data_root + 'annotations/instances_val2017.json',
        img_prefix=data_root + 'val2017/',
        pipeline=[
            dict(type='LoadImageFromFile', to_float32=True),
            dict(type='LoadAnnotations', with_bbox=True)
        ],
        classes=CLASSES,
        filter_empty_gt=False,
        test_mode=True,
        iscrowd=True),
    pipeline=test_pipeline,
    dynamic_scale=None,
    label_padding=False)

data = dict(
    imgs_per_gpu=16, workers_per_gpu=4, train=train_dataset, val=val_dataset)

# additional hooks
interval = 10
custom_hooks = [
    dict(
        type='YOLOXModeSwitchHook',
        no_aug_epochs=15,
        skip_type_keys=('MMMosaic', 'MMRandomAffine', 'MMMixUp'),
        priority=48),
    dict(
        type='SyncRandomSizeHook',
        ratio_range=random_size,
        img_scale=img_scale,
        interval=interval,
        priority=48),
    dict(
        type='SyncNormHook',
        num_last_epochs=15,
        interval=interval,
        priority=48)
]

# evaluation
eval_config = dict(
    interval=10,
    gpu_collect=False,
    visualization_config=dict(
        vis_num=10,
        score_thr=0.5,
    )  # show by TensorboardLoggerHookV2 and WandbLoggerHookV2
)
eval_pipelines = [
    dict(
        mode='test',
        data=data['val'],
        evaluators=[dict(type='CocoDetectionEvaluator', classes=CLASSES)],
    )
]

checkpoint_config = dict(interval=interval)

# optimizer
optimizer = dict(
    type='SGD', lr=0.02, momentum=0.9, weight_decay=5e-4, nesterov=True)
optimizer_config = {}

# learning policy
lr_config = dict(
    policy='YOLOX',
    warmup='exp',
    by_epoch=False,
    warmup_by_epoch=True,
    warmup_ratio=1,
    warmup_iters=5,  # 5 epoch
    num_last_epochs=15,
    min_lr_ratio=0.05)

# exponetial model average
ema = dict(decay=0.9998)

# runtime settings
total_epochs = 300

# yapf:disable
log_config = dict(
    interval=100,
    hooks=[
        dict(type='TextLoggerHook'),
        dict(type='TensorboardLoggerHookV2'),
        # dict(type='WandbLoggerHookV2'),
    ])

export = dict(export_type = 'ori', preprocess_jit = False, batch_size=1, blade_config=dict(enable_fp16=True, fp16_fallback_op_ratio=0.01), use_trt_efficientnms=False)



# Copyright (c) Alibaba, Inc. and its affiliates.
from modelscope.models.builder import MODELS
from modelscope.utils.constant import Tasks

from easycv.models.pose.top_down import TopDown
from easycv.toolkit.modelscope.metainfo import EasyCVModels as Models
from easycv.toolkit.modelscope.models.base import EasyCVBaseModel


@MODELS.register_module(
    group_key=Tasks.human_wholebody_keypoint,
    module_name=Models.human_wholebody_keypoint)
class HumanWholeBodyKeypoint(EasyCVBaseModel, TopDown):

    def __init__(self, model_dir=None, *args, **kwargs):
        EasyCVBaseModel.__init__(self, model_dir, args, kwargs)
        TopDown.__init__(self, *args, **kwargs)
